#!/usr/local/bin/ruby -s

=begin

earth.rb

A mail forwarding control script
Copyright (C) 2001-2003 Masaharu FUJITA

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the author by:
        e-mail: m@fjts.org
        url: http://fjts.org/~m/Soft/Ruby/earth.rb/
=end

Version = "3.6.0"

# Standard library of Ruby
require "net/smtp"
require "net/pop"
require "jcode"
require "kconv"
require "base64"

# Setting of Earth
$KCODE = 'e'
File.umask(63) # umask 077

# Local library of Earth
require "lib/earth_log"
require "lib/earth_func"
require "lib/earth_mail"
require "lib/earth_match"
require "lib/earth_cut"
require "lib/earth_cutline"
require "lib/earth_info"
require "lib/earth_ctrlcmd"
require "lib/earth_suspend"

# Setting file of Earth
require "earth_setting"

# File or directory of Earth
EarthDir = File.dirname($0) # Directory which this script exists
LogDir = EarthDir + '/log' # Log directory
Log = LogDir + '/messages' # Log file
SpoolDir = EarthDir + '/spool' # Spool mail directory
SpoolCount = SpoolDir + '/.number' # Spool mail number
DBDir = EarthDir + '/db' # Database mail directory
MqueueDir = EarthDir + '/mqueue' # Mail queueing directory
MqueueCount = MqueueDir + '/.number' # Mail queueing number
Suspend = EarthDir + '/.suspend' # For Suspend file

# Main class
class Earth
  def initialize
    # check directory and mkdir.
    if FileTest.writable?(EarthDir)
      if not FileTest.exist?(LogDir)
	Dir.mkdir(LogDir)
      end
      
      if not FileTest.exist?(SpoolDir)
	Dir.mkdir(SpoolDir)
      end
      
      if not FileTest.exist?(DBDir)
	Dir.mkdir(DBDir)
      end
      
      if not FileTest.exist?(MqueueDir)
	Dir.mkdir(MqueueDir)
      end
    end # if FileTest.writable?(EarthDir)

    if 'Local' == GetMailType
      get_add_spool_number
    end
  end # def initialize
    
  def get_add_spool_number
    if not File.exist?(SpoolCount)
      @spool_number = 0
    else
      begin
        file = open(SpoolCount, "r")
        file.flock(File::LOCK_EX)
      rescue
        @spool_number = 0
      else
        @spool_number = file.gets.to_i + 1
        if @spool_number >= SpoolNumberMax
          @spool_number = 0
        end
        file.flock(File::LOCK_UN)
        file.close
      end
    end # if not File.exist(SpoolCount)
    
    begin
      file = open(SpoolCount, "w")
      file.flock(File::LOCK_EX)
    rescue
    else
      file.print @spool_number
      file.flock(File::LOCK_UN)
      file.close
    ensure
    end
  end # def get_add_spool_number

  def reverse_spool_number
    if @spool_number > 0
      @spool_number -= 1
    else
      @spool_number = SpoolNumberMax
    end

    begin
      file = open(SpoolCount, "w")
      file.flock(File::LOCK_EX)
    rescue
    else
      file.print @spool_number
      file.flock(File::LOCK_UN)
      file.close
    ensure
    end
  end # def reverse_spool_number

  def get_info_and_match(obj)
    @mail = EarthGetMailInfo.new(MaxHeaderLineOfOneMail,
				MaxBodyByteOfOneMail,MaxAnalsysAddressNum,
				MaxInfoAddressNum,obj)

    if 'POP' == GetMailType and ( /.*O/ =~ @mail.header['status'] or
				 /.*U/ =~ @mail.header['status'] )
      @nowrite = true
      @log.set_end
      return
    end
    
    if UseSpool and not $nospool
      # Save mail.
      mailfilepath = SpoolDir + '/' + @spool_number.to_s
      begin
	@mailfile = open(mailfilepath, "w")
      rescue
      else
	# Write in mail at @mailfile.
	@mail.raw_mail.each do |line|
	  @mailfile.print line
	end
	@mailfile.close
      end # begin
    end # if UseSpool and not $nospool
    
    @match = EarthRuleMatch.new(Rule,@mail.header,@mail.all_header,@mail.from,
			       @mail.to,@mail.cc,@mail.tocc,DefaultCutMaxByte,
			       DefaultMaxNumber,DefaultRegetMaxByte,
			       DefaultRegetMaxNumber)
    @match.judge

    @cut_max_byte = @match.cut_max_byte
    @cut_max_num = @match.cut_max_num
  end # def get_info_and_match

  def do_ctrlcmd
    ctrlcmd = EarthCtrlCmd.new(@mail.header,@mail.body,@mail.from,@mail.to,
			       CtrlCmdPassword,CtrlCmdReceived,
			       CtrlCmdMessageID,CtrlCmdFromList,CtrlCmdTo)
    if ctrlcmd.check
      case ctrlcmd.type
      when 'reget'

	@log.write_first
	@log.write_info(@spool_number,@mail.info_from,@mail.subject,
			@mail.info_tocc,@mail.date,nil)

	if not UseSpool
	  @log.write_error("Don\'t exec reget that you must set UseSpool = true.")
	  @log.write_last
	  @log.set_end
	  return
	end
	
	if 'just' == ctrlcmd.rest.strip
	  $just = true
	end

	$noinfo = true
	$nodb = true

	ctrlcmd.seq.each do |i|
	  @log.write_reget(i)

	  filename = SpoolDir + '/' + i.to_s
	  @spool_number = i
	  
	  # Getting spool file
          get_spool_file_err = true
          if File.exist?(filename)
  	    begin
	      file = open(filename, "r")
	    rescue
	    else
	      filebody = file.readlines
	      file.close
              get_spool_file_err = false
	    end
          end # if not File.exist?(filename)

	  if get_spool_file_err
	    @log.write_error("Don\'t exec reget with probrem of " + filename +
			     ".")
            next
          end
	  
	  @mail = EarthGetMailInfo.new(MaxHeaderLineOfOneMail,
				       MaxBodyByteOfOneMail,
				       MaxAnalsysAddressNum,
				       MaxInfoAddressNum, filebody)
	  @match = EarthRuleMatch.new(Rule,@mail.header,@mail.all_header,
				      @mail.from,@mail.to,@mail.cc,
				      @mail.tocc,DefaultCutMaxByte,
				      DefaultMaxNumber,DefaultRegetMaxByte,
				      DefaultRegetMaxNumber)
	  @match.judge
	  
	  @cut_max_byte = @match.reget_cut_max_byte
	  @cut_max_num = @match.reget_cut_max_num
	  
	  cut_and_send
	end # ctrlcmd.seq do |i|

	@log.set_end
	return
      when 'send'

	@log.write_first
	@log.write_info(@spool_number,@mail.info_from,@mail.subject,
			@mail.info_tocc,@mail.date,nil)

	create_header = EarthCreateHeader.new(VicariousFrom,
					      VicariousFrom,
					      ctrlcmd.seq.join(','))
	create_header.create
	to = ctrlcmd.seq + VicariousBcc
	body = ctrlcmd.rest.strip + VicariousSignature

	if Regexp.new(SubjectPause) =~ body
	  subject, body = body.split(SubjectPause,2)
	  print(subject)
	  print(body)
	  print("\n")
	else
	  subject = VicariousSubject
	end

	if $mail
	  mailformat = EarthMailFormat.new(create_header.header,
					   subject,body)
	  sendmail = EarthSendMail.new(SmtpServer,SmtpPort,mailformat.mail,
				       VicariousFrom,to,
				       SmtpUser,SmtpPasswd,SmtpAuthType,
				       UseSmtpAuth)
	  sendmail.send
	else
	  mail_print = EarthPrint.new(create_header.header,
				      subject,body)
	  mail_print.header_tag
	  mail_print.header
	  mail_print.body_tag
	  mail_print.body
	end

	ctrlcmdlog = "@log.write_send(to)"
      when 'reply'
	@log.write_first
	@log.write_info(@spool_number,@mail.info_from,@mail.subject,
			@mail.info_tocc,@mail.date,nil)

	if not UseSpool
	  @log.write_error("Don\'t exec reply with probrem of " + filename +
			   ".")
	  @log.write_last
	  @log.set_end
	  return
	end
	
        reply_mail_number = ctrlcmd.seq[0]
	filename = SpoolDir + '/' + reply_mail_number.to_s

	# Getting spool file
        if not File.exist?(filename)
	  return
        else
	  begin
	    file = open(filename, "r")
	  rescue
	    return
	  else
	    filebody = file.readlines
	    file.close
	  end
        end

	@mail = EarthGetMailInfo.new(MaxHeaderLineOfOneMail,
				     MaxBodyByteOfOneMail,
				     MaxAnalsysAddressNum,
				     MaxInfoAddressNum, filebody)

	create_header = EarthCreateHeader.new(VicariousFrom,
					      VicariousFrom,
					      @mail.from,
					      @mail.header['message-id'],
					      @mail.header['references'])
	create_header.create

	to = [@mail.from] + VicariousBcc
	body = ctrlcmd.rest.strip + VicariousSignature
	subject = 'Re: ' + @mail.subject.gsub(/Re:/,'').gsub(/RE:/,'')

	if $mail
	  mailformat = EarthMailFormat.new(create_header.header,
					   subject,body)
	  sendmail = EarthSendMail.new(SmtpServer,SmtpPort,mailformat.mail,
				       VicariousFrom,to,
				       SmtpUser,SmtpPasswd,SmtpAuthType,
				       UseSmtpAuth)
	  sendmail.send
	else
	  mail_print = EarthPrint.new(create_header.header,
				      subject,body)
	  mail_print.header_tag
	  mail_print.header
	  mail_print.body_tag
	  mail_print.body
	end

	ctrlcmdlog = "@log.write_reply(reply_mail_number)"
      end # case ctrlcmd.type

      eval(ctrlcmdlog)

      @log.set_end
      return
    elsif ctrlcmd.error
      @log.write_first
      @log.write_info(@spool_number,@mail.info_from,@mail.subject,
		      @mail.info_tocc,@mail.date,nil)

      messages = "Don\'t exec control command. " + ctrlcmd.error
      messages += " is not match."

      @log.write_error(messages)

      create_header = EarthCreateHeader.new(ReturnPathAddress,
                                            ReturnPathAddress,To)
      create_header.create
      subject = 'Error: ' + @spool_number.to_s

      if $mail
	mailformat = EarthMailFormat.new(create_header.header,
					 subject, messages)
	sendmail = EarthSendMail.new(SmtpServer,SmtpPort,mailformat.mail,
				     VicariousFrom,To,
				     SmtpUser,SmtpPasswd,SmtpAuthType,
				     UseSmtpAuth)
	sendmail.send
      else
	mail_print = EarthPrint.new(create_header.header,
				    subject,messages)
	mail_print.header_tag
	mail_print.header
	mail_print.body_tag
	mail_print.body
      end

      @log.set_end
      @nowrite = true
      return
    end # if ctrlcmd.check
  end # def do_ctrlcmd

  def send_infomation(mail_number)
    create_info_header = EarthCreateHeader.new(ReturnPathAddress,
					       InfoMailFromAddress,To)
    create_info_header.create
    
    if UseSubject
      info_mail_subject = EarthFormat.new(PhoneType,InfoMailSubject,
					  @mail.info_from,@mail.subject,
					  @mail.info_tocc,@mail.date,
					  mail_number,0, @match.match,
					  SubjectMaxByte)
      info_mail_subject.create
      subject_in_format = info_mail_subject.format
    else
      subject_in_format = ''
    end # if UseSubject
    
    info_mail_body = EarthFormat.new(PhoneType,InfoMailBody,@mail.info_from,
				     @mail.subject,@mail.info_tocc,
				     @mail.date,mail_number,0,@match.match,
				     -1,false)
    info_mail_body.create
    
    if UseSuspend and 
	( @suspend.stats == 'suspend' or @suspend.stats == 'restart' )
      mailformat = EarthMailFormat.new(create_info_header.header,
				       subject_in_format,
				       info_mail_body.format)
      @index = 0
      @suspend.save(mailformat.mail,@index)
      @index += 1
      @log.write_info_mail_in_suspend(mail_number)
    else
      if $mail
	mailformat = EarthMailFormat.new(create_info_header.header,
					 subject_in_format,
					 info_mail_body.format)
	sendmail = EarthSendMail.new(SmtpServer,SmtpPort,mailformat.mail,
				     ReturnPathAddress,To,
				     SmtpUser,SmtpPasswd,SmtpAuthType,
				     UseSmtpAuth)
	sendmail.send
      else
	subject_byte = subject_in_format.length
	from_byte = InfoMailFromAddress.length
	info_byte = info_mail_body.byte + eval(MailByteContent)
	
	info_print = EarthPrint.new(create_info_header.header,
				    subject_in_format,
				    info_mail_body.format,
				    info_byte)
	info_print.header_tag
	info_print.header
	info_print.body_tag
	info_print.body
      end
      
      @log.write_info_mail(mail_number)
    end # if UseSuspend and 
  end # def send_infomation
  
  def cut_and_send
    @log.write_first
    @log.write_info(@spool_number,@mail.info_from,@mail.subject,
		    @mail.info_tocc,@mail.date,@match.match)

    if $just
      if FromType
	send_header_from = FromType
      else
	send_header_from = @mail.from
      end
      create_header = EarthCreateHeader.new(ReturnPathAddress,
					    send_header_from,To)
      create_header.create

      if $mail
	mailformat = EarthMailFormat.new(create_header.header,
					 @mail.subject,@mail.body)
	
	sendmail = EarthSendMail.new(SmtpServer,SmtpPort,mailformat.mail,
				     ReturnPathAddress,To,
				     SmtpUser,SmtpPasswd,SmtpAuthType,
				     UseSmtpAuth)
	sendmail.send
      else
	header_print = EarthPrint.new(create_header.header)
	header_print.header_tag
	header_print.header
	header_print.body_tag
	mail_print = EarthPrint.new(nil,@mail.subject,@mail.body)
				    
	mail_print.body
      end

      @log.write_normal(@spool_number, nil, true)
      @log.set_end
      return
    end # if $just

    if @cut_max_byte == 0 or @cut_max_num == 0
      @log.write_no_send(@spool_number)
      @log.set_end
      return
    end

    mail_number = @spool_number

    if SendHeaderInfo and not $noinfo and not HeaderInfoLastSend
      send_infomation(mail_number)
    end # if SendHeaderInfo and not $noinfo and not HeaderInfoLastSend

    body = multi2ascii(@mail.body)
    # from string to array
    body_array = body.split("\n")

    if UseDB and not $nodb
      db = EarthDB.new(@mail.from,body_array,DBMaxFileByte,DBDir)
      db.do
      body_array = db.body
    end

    if CutLine
      cutline = EarthCutLine.new(CutLinePattern,body_array)
      cutline.do
      body_array = cutline.body
    end

    if EditReplyLine
      replyline = EarthEditReplyLine.new(WriteOneReplyLine,body_array)
      replyline.do
      body_array = replyline.body
    end

    # from array to string
    body = body_array.join("\n")

    @linefeedcharacter = "\n"
    if ChangeLineFeed
      body = body.gsub(/\n\n+/,LineFeedCharacter).gsub(/\n/,' ')
      @linefeedcharacter = LineFeedCharacter
    end
    
    if UseSqueeze
      body = special_squeeze(body,SqueezeCharacter)
    end

    if not $nobody and body != nil and body != ''
      if $mail and not ( UseSuspend and       
			( @suspend.stats == 'suspend' or 
			 @suspend.stats == 'restart' ))
	sleep LatencyTime
      end

      if FromType
	send_header_from = FromType
      else
	send_header_from = @mail.from
      end
      create_header = EarthCreateHeader.new(ReturnPathAddress,
					    send_header_from,To)
      create_header.create
      
      if not $mail and not ( UseSuspend and       
			    ( @suspend.stats == 'suspend' or 
			     @suspend.stats == 'restart' ))
	header_print = EarthPrint.new(create_header.header)
	header_print.header_tag
	header_print.header
	header_print.body_tag
      end

      @count = 0
      @is_first = true
      init_subject_and_head
      while body
	str = EarthCutString.new(PhoneType,body,@cut_byte,@linefeedcharacter)
	
	if UseSuspend and 
	    ( @suspend.stats == 'suspend' or @suspend.stats == 'restart' )
	  mailformat = EarthMailFormat.new(create_header.header,
					   @subject_in_cutbody,
					   @mail_head.format + str.string)
	  @suspend.save(mailformat.mail,@index)
	  @index += 1
	else
	  if $mail
	    mailformat = EarthMailFormat.new(create_header.header,
					     @subject_in_cutbody,
					     @mail_head.format + str.string)

	    sendmail = EarthSendMail.new(SmtpServer,SmtpPort,mailformat.mail,
	    				 ReturnPathAddress,To,
					 SmtpUser,SmtpPasswd,SmtpAuthType,
					 UseSmtpAuth)
	    sendmail.send
	  else
	    mail_print = EarthPrint.new(nil,@subject_in_cutbody,
					@mail_head.format + str.string,
					str.byte + @etc_byte,@count)
	    mail_print.body
	  end
	end # if UseSuspend and
	
	body = str.rest
	@count += 1
	@is_first = false
	if @cut_max_num > 0 and @count >= @cut_max_num
	  break
	end
	init_subject_and_head
	

	if $mail and not ( UseSuspend and       
			  ( @suspend.stats == 'suspend' or 
			   @suspend.stats == 'restart' ))
	  sleep LatencyTime
	end
      end # while body

      if UseSuspend and
	  ( @suspend.stats == 'restart' or @suspend.stats == 'suspend')
	@log.write_suspend(mail_number,@count)
      else
	@log.write_normal(mail_number,@count)
      end
    end # if not $nobody

    if SendHeaderInfo and not $noinfo and HeaderInfoLastSend
      send_infomation(mail_number)
    end # if SendHeaderInfo and not $noinfo and HeaderInfoLastSend
    @log.set_end
  end # def cut_and_send

  def init_subject_and_head
    if UseSubject
      @mail_subject = EarthFormat.new(PhoneType,SubjectFormat,
				      @mail.info_from,@mail.subject,
				      @mail.info_tocc,@mail.date,@spool_number,
				      @count,@match.match, SubjectMaxByte)
      @mail_subject.create
      @subject_in_cutbody = @mail_subject.format
      subject_byte =  @mail_subject.byte
    else
      @subject_in_cutbody = ''
      subject_byte =  0
    end # if UseSubject
    
    if @is_first
      head_format = FirstBodyHead
    else
      head_format = NextBodyHead
    end
    
    @mail_head = EarthFormat.new(PhoneType,head_format,@mail.info_from,
				 @mail.subject,@mail.info_tocc,@mail.date,
				 @spool_number,@count)
    @mail_head.create
    
    from_byte = @mail.from.length
    @etc_byte = @mail_head.byte + eval(MailByteContent)
    
    @cut_byte = @cut_max_byte - @etc_byte
  end # def init_subject_and_head

  def do_restart
    if FileTest.exist?(MqueueCount)
      File.delete(MqueueCount)
    end

    @log.write_restart
    
    i = 0
    j = 0
    while i <= @suspend.count
      filename = MqueueDir + '/' + i.to_s + '-' + j.to_s

      if not FileTest.exist?(filename)
	@log.write_normal_in_restart(i,j)

	i += 1
	j = 0
	next
      end
      
      # Get mail byte.
      get_mail_byte_err = true
      if File.exist?(filename)
        begin
          file = open(filename,"r")
        rescue
        else
          get_mail_byte_err = false
  	  filebody = file.readlines

	  if $mail
	    sendmail = EarthSendMail.new(SmtpServer,SmtpPort,filebody,
	  			         ReturnPathAddress,To,
				         SmtpUser,SmtpPasswd,SmtpAuthType,
				         UseSmtpAuth)
	    sendmail.send
	    sleep LatencyTime
	  else
	    mail_print = EarthPrint.new(filebody)
	    mail_print.header_tag
	    mail_print.header
	  end

          file.close
	  File.delete(filename)
        ensure
        end
      end

      if get_mail_byte_err
	next
      end

      j += 1
    end # while i <= @suspend.count
  end # def do_restart

  def check_suspend
    @suspend = EarthSuspend.new(@log.time,Holiday,WeekdaySuspendTime,
				HolidaySuspendTime,Suspend,
				MqueueDir,MqueueCount)
    @suspend.check
    
    if @suspend.stats == 'suspend'
      suspend_match = EarthRuleMatch.new(IgnoreSuspendRule,
					 @mail.header,@mail.all_header,
					 @mail.from,@mail.to,@mail.cc,
					 @mail.tocc)
      suspend_match.judge

      if suspend_match.match
	@suspend.stats_change('normal')
      end
    end # if @suspend.stats = 'suspend'
  end # def check_suspend

  def action(obj)
    @nowrite = false

    get_info_and_match(obj)
    if @log.is_end
      if not @nowrite
	@log.write_end
      end

      return
    end

    if UseSuspend
      check_suspend
      if @log.is_end
	@log.write_end
	return
      end
    end

    if UseCtrlCmd
      do_ctrlcmd
      if @log.is_end
	@log.write_end
	return
      end
    end # if UseCtrlCmd

    cut_and_send

    if UseSuspend and @suspend.stats == 'restart'
      do_restart
    end # if UseSuspend and @suspend.stats == 'restart'

    @log.write_end
  end # def action

  def main
    @log = EarthLog.new(Log,SpoolNumberMax)

    case GetMailType
    when 'Local'
      action($stdin)
      exit
    when 'POP'
      PopSettings.each do |parts|
	pop_type = parts[0]
	server = parts[1]
	user = parts[2]
	passwd = parts[3]
	if parts[4] == nil
	  port = 110
	else
	  port = parts[4]
	end

	case pop_type
	when 'POP'
	  p = Net::POP3.new(server,port)
	when 'APOP'
	  p = Net::APOP.new(server,port)
	end # case pop_type

	begin
	  p.start(user,passwd)
	rescue
	  @log.write_pop_error(pop_type,server,user,port)
	  next
	else
	end
	@log.write_pop_start(pop_type,server,user,port)
	
	if not p.mails.empty?
	  p.each_mail do |m|
	    get_add_spool_number
	    action(m.pop.split("\n"))
	    if @nowrite
	      reverse_spool_number
	    end
	    @log.init
	  end # p.each_mail do |m|
	end # if not p.mails.empty?
	
	p.reset
	p.finish
	@log.write_pop_end(pop_type,server,user,port)
      end #PopSettings.each do |parts|
    end # case @type
  end # def main
end # class Earth

###### Main program ######
if $help
  print "usage: #$0 [-nospool] [-nolog] [-nodb] [-noinfo] [-nobody] [-just] [-mail]\n"
  exit
end

earth = Earth.new
earth.main
###### Main program ######
