=begin

earth_cutline.rb

Local mail library for a mail forwarding control script
Copyright (C) 2001-2003 Masaharu FUJITA

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the author by:
        e-mail: m@fjts.org
        url: http://fjts.org/~m/Soft/Ruby/earth.rb/
=end

# class for cutting line for database of from address
class EarthDB
  def initialize(from,mail_body,max_file_byte,dbdir)
    @from = from
    @mail_body = mail_body
    @max_file_byte = max_file_byte
    @dbdir = dbdir

    @return_mail_body = Array.new
  end # def initialize

  def do
    if @from == nil or @mail_body == nil
      return
    end

    # Setting filename for database.
    filename = @dbdir + '/' + @from
    backupbody = @mail_body

    # Getting database.
    database = nil
    if File.exist?(filename)
      begin
        file = open(filename, "r")
        file.flock(File::LOCK_EX)
      rescue
      else
        database = file.readlines
        file.flock(File::LOCK_UN)
        file.close
      end
    end

    # Save database file.
    begin
      file = open(filename, "w")
      file.flock(File::LOCK_UN)
    rescue
    else
      backupbody.each do |line|
        file.print line
	file.print "\n"
      end

      if database != nil
        database.each do |line|
          file.print line
        end
      end

      file.flock(File::LOCK_UN)
      file.close
    ensure
    end

    # Adjustment of file size.
    filesize = File.size(filename)
    maxfilesize = @max_file_byte
    if filesize != nil and filesize > maxfilesize
      begin
        File.truncate(filename,maxfilesize)
      rescue
      else
      end
    end # if filesize != nil and filesize > maxfilesize

    if database == nil
      @return_mail_body = @mail_body
      return
    end

    # Cut line with database
    before_match = false
    is_match = false
    tmp = nil
    mail_body_idx = 0
    database_idx = 0

    while @mail_body[mail_body_idx] != nil
      if "\n" == @mail_body[mail_body_idx]
	@return_mail_body.push(@mail_body[mail_body_idx])
	mail_body_idx += 1
	next
      end

      if not is_match
	database_idx = 0
      end

      while database[database_idx] != nil
	if not @mail_body[mail_body_idx]
	  break
	end

	if not is_match and "\n" == database[database_idx]
	  database_idx += 1
	  next
	end

	before_match = is_match

	if database[database_idx].strip == @mail_body[mail_body_idx].strip
	  if not before_match
	    tmp = @mail_body[mail_body_idx]
	  end

	  is_match = true
	  database_idx += 1
	  mail_body_idx += 1
	  next
	end # if database[database_idx] == @mail_body[mail_body_idx]

	if before_match
	  database_idx += 1
	  break
	end

	is_match = false
	database_idx += 1
      end # while database[database_idx] != nil

      if not is_match and before_match
	@return_mail_body.push(tmp)
      end

      @return_mail_body.push(@mail_body[mail_body_idx])
      mail_body_idx += 1
    end # while @mail_body[mail_body_idx] != nil
  end # def do

  def body
    @return_mail_body
  end
end # class EarthDB

# class for cutting line some pattern
class EarthCutLine
  def initialize(pattern,mail_body)
    @pattern = pattern
    @mail_body = mail_body
    @return_mail_body = Array.new
  end # def initialize(pattern,mail_body)

  def do
    @mail_body.each do |line|
      is_cut = false
      if not line
	break
      end

      @pattern.each do |p|
	if line.strip =~ Regexp.new(p)
	  is_cut = true
	  break
	end
      end # @pattern.each do |p|

      if is_cut
	next
      end

      @return_mail_body.push(line)
    end # @mail_body.each do |line|
  end # def do

  def body
    @return_mail_body
  end
end # class EarthCutLine

# class for edit reply line.
class EarthEditReplyLine
  def initialize(write_one_line,mail_body)
    @write_one_line = write_one_line
    @mail_body = mail_body
    @return_mail_body = Array.new
  end # def initialize(edit,write_one_line,mail_body)

  def do
    is_reply = false
    @mail_body.each do |line|
      if ( not /.*@.*>/ =~ line ) and /.*> .*/ =~ line or /^[ ]*[:|>]/ =~ line
	if is_reply
	  next
	end

	is_reply = true
	@return_mail_body.push("\n\n\n")

	if not @write_one_line
	  next
	end
      else
	if is_reply
	  @return_mail_body.push("\n\n\n")
	end

	is_reply = false
      end # if ( not /.*@.*>/ =~ line ) and .......

      @return_mail_body.push(line)
    end # @mail_body.each do |line|
  end # def do

  def body
    @return_mail_body
  end
end # class EarthEditReplyLine
