=begin

earth_mail.rb

Local mail library for a mail forwarding control script
Copyright (C) 2002-2003 Masaharu FUJITA

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You may contact the author by:
        e-mail: m@fjts.org
        url: http://fjts.org/~m/Soft/Ruby/earth.rb/
=end

# for get mail infomation.
class EarthGetMailInfo
  def initialize(max_line, max_byte, max_analsys_address_num,
		 max_info_address_num, mail_obj)
    @max_line = max_line
    @max_byte = max_byte
    @max_analsys_address_num = max_analsys_address_num
    @max_info_address_num = max_info_address_num
    @mail_obj = mail_obj

    @raw_mail = Array.new
    @raw_mail_idx = 0

    get_mail(@mail_obj)
    create_info
  end # def initialize

  # get mail
  def get_mail(obj)
    is_header = true

    @header = Hash.new
    header_line = 0
    key = nil
    value = nil
    boundary = nil

    is_multipart = false
    @body = ''
    body_byte = 0
    html_body = ''
    is_body_header = false
    content_type = nil
    mp_body = Array.new
    mp_index = 0

    if obj == $stdin
      get_line_string = "line = obj.gets"
    else
      get_line_string = "line = obj.shift"
    end

    # Get mail
    while eval(get_line_string) and body_byte <= @max_byte
      # line convertion.
      line = Kconv.toeuc(line.gsub(/\r\n/,"\n").gsub(/\r/,"\n"))
      @raw_mail.push(line)

      if is_header # Header case
	header_line += 1

	if "\n" == line or header_line >= @max_line
	  is_header = false
	  if /multipart/i =~ @header['content-type']
	    is_multipart = true
	    tmp, boundary = @header['content-type'].split(/boundary="/i,2);
	    if boundary != nil
	      boundary, tmp = boundary.split(/"/,2);
	      boundary = '--' + boundary
            else
	      boundary = '--'
            end
	  end

	  if "\n" == line
	    next
	  end
	end # if "\n" == line or header_line >= @max_line

	# Create header element
	if /^[\t ]+/ =~ line
	  value = @header[key] + ' ' + decode_b(line.strip)
	else
	  key, value = line.split(/: /,2)
	  key = key.downcase
	  if @header.key?(key)
	    value = @header[key] + ' ' + decode_b(value).strip
	  end
	end

        if value == nil
          value = ''
        end

	@header[key] = decode_b(value).strip
      else # Body case
	if not is_multipart
	  case @header['content-type']
	  when /text\/plain/i 
	    @body += line
	    body_byte += line.length
	  when /text\/html/i
	    html_body += line
	    body_byte += line.length
	  end # case content_type
	  
	  next
	end

	# Multipart case
	if boundary + '--' == line.strip or boundary == line.strip
	  case content_type
	  when /message\/rfc822/i
	    tmp = EarthGetMailInfo.new(@max_line,@max_byte,
				       @max_analsys_address_num,
				       @max_info_address_num,mp_body)
	    @body += tmp.body
	    body_byte += tmp.body.length
	  when /text\/html/i
	    @body += html2text(html_body)
	    html_body = ''
	  end # case content_type
	  
	  if @body != ''
	    @body += "\n\n\n"
	  end

	  is_body_header = true
	  next
	end

	if is_body_header
	  if /content-type/i =~ line
	    tmp, content_type = line.split(': ',2);
	  elsif "\n" == line
	    is_body_header = false
	    mp_body = Array.new
	    mp_index = 0
	  end
	  next
	end # if is_body_header

	case content_type
	when /text\/plain/i 
	  @body += line
	  body_byte += line.length
	when /text\/html/i
	  html_body += line
	  body_byte += line.length
	when /message\/rfc822/i
	  mp_body[mp_index] = line
	  mp_index += 1
	end # case content_type
      end # if is_header
    end # while line = $stdin.readline or body_byte <= @max_byte

    @body += html2text(html_body)
  end # def get_mail

  def create_info
    # create infomation
    @from = get_address(@header['from'])

    @to = get_address(@header['to'])
    @cc = get_address(@header['cc'])
    @tocc = nil
    if @to
      @tocc = @to
    end
    if @cc
      if @tocc
	@tocc += ',' + @cc
      else
	@tocc = @cc
      end
    end

    @info_from = create_info_address(@from)
    @info_to = create_info_address(@to)
    @info_cc = create_info_address(@cc)
    @info_tocc = create_info_address(@tocc)

    @date = nil    
    if @header.key?('date')
      tmp, day, month, year, time, tmp = @header['date'].split
      if tmp != nil
        hour, min, sec = time.split(':')
        @date =
  	  Time.local(year,month,day,hour,min,sec).strftime("%Y/%m/%d %H:%M")
      end
    end

    @all_header = @header.values.join(' ')
  end # def create_info

  # for get e-mail address
  def get_address(line)
    if not line
      return nil
    end

    # Check of ""
    array = Array.new
    tmp_string = ''
    is_text = false
    tmp_array = line.split(',')
    number = 0

    tmp_array.each do |tmp|
      if /.*\".*/ =~ tmp and not /.*\".*\".*/ =~ tmp
	tmp_string = tmp_string + "," + tmp
	is_text = ( not is_text )
      elsif is_text
	tmp_string = tmp_string + "," + tmp
      else
	tmp_string = tmp
      end

      if not is_text
	array.push(tmp_string)
	tmp_string = ''
	number += 1
      end

      if number >= @max_analsys_address_num
	break
      end
    end # tmp_array.each do |tmp|

    # Check of ()
    tmp_array = array
    array = Array.new
    tmp_string = ''
    is_text = false

    tmp_array.each do |tmp|
      if /.*\(.*\).*/ =~ tmp
	array.push(tmp)
	next
      end

      if /.*\(.*/ =~ tmp
	tmp_string = tmp_string + "," + tmp
	is_text = true
      elsif is_text
	if /.*\).*/ =~ tmp
	  is_text = false
	end
	tmp_string = tmp_string + "," + tmp
      else
	tmp_string = tmp
      end

      if not is_text
	array.push(tmp_string)
	tmp_string = ''
      end
    end #tmp_array.each do |tmp|

    # get real mail address
    address = ''
    is_address = false
    array.each do |sub|
      orig = sub.strip 
      str = orig

      if not /.*[\"()].*/ =~ str
	is_address,str = check_address(str)
	if not is_address
	  str = orig
	end
      else
	# Grouping
	#  normal_entry ( comment_entry ) normal_entry
	#  normal_entry " comment_entry " normal_entry
	normal_entry = Array.new
	comment_entry = Array.new

	is_d = false
	is_k = false
	end_of_entry = false
	is_normal_entry = true
	entry_string = ''

	str.each_char do |c|
	  case c
	  when '"'
	    if is_k
	      entry_string += c
	    else
	      is_d = ( not is_d )
	      end_of_entry = true
	    end
	  when '('
	    if is_d or is_k
	      entry_string += c
	    else
	      is_k = true
	      end_of_entry = true
	    end
	  when ')'
	    if not is_d and is_k
	      end_of_entry = true
	      is_k = false
	    else
	      entry_string += c
	    end
	  else
	    entry_string += c
	  end
	  
	  if not end_of_entry
	    next
	  end
	  
	  if is_normal_entry
	    normal_entry.push(entry_string)
	  else
	    comment_entry.push(entry_string)
	  end
	  end_of_entry = false
	  entry_string = ''
	  is_normal_entry = ( not is_normal_entry )
	end # str.each_char do |c|
	
	if is_normal_entry
	  normal_entry.push(entry_string)
	else
	  comment_entry.push(entry_string)
	end

	# Check mail address in normail_entry.
	is_address = false
	normal_entry.each do |p|
	  is_address,str = check_address(p)
	  if is_address
	    break
	  end
	end
	
	# Check mail address in comment_entry.
	if not is_address
	  comment_entry.each do |p|
	    is_address,str = check_address(p)
	    if is_address
	      break
	    end
	  end
	end # if not is_address
	
	# Invalid mail address format.
	if not is_address
	  str = orig
	end
      end # if /.*[\"()].*/ =~ str
      
      if not address.empty?
	address.concat(',')
      end
      address.concat(str)
    end # array.each do |sub|
 
    return address
  end # def get_address(line)

  def check_address(str)
    is_address = false
    if /.*<.*>.*/ =~ str
      tmp, str = str.split('<',2)
      str, tmp = str.split('>',2)
      is_address = true
    elsif /.*@.*/ =~ str
      is_address = true
    end

    return is_address, str
  end #def check_address(str)


  # create address for use infomation.
  def create_info_address(address)
    if not address
      return nil
    end

    array = address.split(',')
    info_address = ''
    number = 0
    array.each do |one|
      if number != 0
        info_address += ','
      end
      info_address += one
      number += 1

      if number >= @max_info_address_num
	break
      end
    end # array.each do |one|

    return info_address
  end # def create_info_address(address)

  # Currently, eliminate tags.
  def html2text(html_string)
    if not html_string
      return ''
    end

    is_tag = false
    text_string = ''
    html_string.each_char do |char|
      case char
      when '<'
	is_tag = true
	next
      when '>'
	if is_tag
	  is_tag = false
	  next
	end
      end
      
      if not is_tag
	text_string += char
      end
    end

    return text_string
  end # def html2text(html_string)

  def from
    @from
  end

  def to
    @to
  end

  def cc
    @cc
  end

  def tocc
    @tocc
  end

  def info_from
    @info_from
  end

  def info_to
    @info_to
  end

  def info_cc
    @info_cc
  end

  def info_tocc
    @info_tocc
  end

  def date
    @date
  end

  def all_header
    @all_header
  end

  def header
    @header
  end

  def body
    @body
  end

  def subject
    header['subject']
  end

  def raw_mail
    @raw_mail
  end
end # class EarthGetMailInfo

# for create mail header
class EarthCreateHeader
  def initialize(return_path,from,to,message_id=nil,references=nil)
    @return_path = return_path
    @from = from
    @to = to
    @message_id = message_id
    @references = references

    @header = Array.new
  end # def initialize

  def create
    @header.push("Return-Path: <" + @return_path + ">\r\n")
    @header.push("From: " + @from + "\r\n")
    @header.push("To: " + @to + "\r\n")
    @header.push("MIME-Version: 1.0\r\n")
    @header.push("Content-Transfer-Encoding: 7bit\r\n")
    @header.push("Content-Type: Text/Plain; charset=iso-2022-jp\r\n")
    @header.push("X-Mailer: earth.rb version " + Version + "\r\n")
    # Correspondence to thread.
    if @message_id
      @header.push("In-Reply-To: " + @message_id + "\r\n")
      
      if not @references
	@header.push("References: " + @message_id + "\r\n")
      else
	array = (@references + " " + @message_id).split
	first = array.shift
	@header.push("References: " + first + "\r\n")
	array.each do |p|
		@header.push("\t" + p + "\r\n")
	end
      end
    end
  end # def create

  def header
    @header
  end
end # class EarthCreateHeader

# for send mail
class EarthSendMail
  def initialize(host,port,mail,from,to,user=nil,passwd=nil,
		 auth_type=nil,use_smtp_auth=false)
    @host = host
    @port = port
    @mail = mail
    @from = from
    @to = to

    if use_smtp_auth
      @user = user
      @passwd = passwd
      @auth_type = auth_type
    else
      @user = nil
      @passwd = nil
      @auth_type = nil
    end
  end # def initialize

  def send
    Net::SMTP.start( @host, @port, Socket.gethostname, @user ,@passwd ,@auth_type ) do |smtp|
      smtp.ready( @from, @to ) do |adapter|
	@mail.each do |line|
	  adapter.write line
	end
      end # smtp.ready
    end # Net::SMTP.start( @host, @port ) do |smtp|
  end # def send
end # class EarthSendMail

class EarthMailFormat
  def initialize(header,subject,body)
    @mail = Array.new
    @mail += header
    @subject = subject

    if @subject and @subject != ''
      is_re = false
      if /^Re: / =~ @subject
	is_re = true
	@subject = @subject.sub(/Re: /,'')
      end

      tmp_subject = @subject
      @subject = '=?ISO-2022-JP?B?'
      @subject += encode64(Kconv.tojis(tmp_subject)).delete("\n").strip
      @subject += '?='

      if is_re
	@subject = 'Re: ' + @subject
      end

      @mail.push("Subject: " + @subject + "\r\n")
    end # if not @subject

    @mail.push("\r\n")
    @mail.push(Kconv.tojis(body).gsub(/\n/,"\r\n"))
  end # def initialize(header,subject,body)

  def mail 
    @mail
  end
end # class EarthMailFormat
